#!/usr/bin/perl -w

# =============================================================================
# csvdiff.pl - compare two csv files, Copyrigth 2006 - today - Roland Schmitz
# =============================================================================
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation version 2.1 of the License.
#
# This program is distributed because i hope that it will be usefull, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License version
# 2.1 for more details.
# =============================================================================
# csvdiff is a Perl script to diff/compare two csv files with the possibility
# to select the separator. Differences will be shown like: "Column XYZ in
# record 999" is different. After this, the actual and the expected result for
# this column will be shown.
# =============================================================================
# Contact: r-sch AT users DOT sourceforge DOT net
# =============================================================================
# Version 2.0  - created: 2010-11-01 by Roland Schmitz, did some Perltidy
#                                       and Perlcritic + feedback from
#                                       Jeremy Courter: use quotes within data
#                                       Thomas Fahle: reduce complexity
#                changed: YYYY-MM-DD by xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
#            last change: YYYY-MM-DD by xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# =============================================================================

#perl csvdiff.pl -a act.csv -e exp.csv -s ";" -c col_names.csv -k "2" |more
#perl -d:ptkdb csvdiff.pl -a act.csv -e exp.csv -s ";" -c col_names.csv -k "2"

use strict;
use warnings;

use Getopt::Std;
use Data::Dumper;
use Scalar::Util qw[ looks_like_number ];
use Carp;

use lib ".";
use csvdiff_function qw(usage buggy chk_input_file read_csv parse_line
  trim trim_delete_space create_hash sort_before_compare cmp_data
  diff_pos call_diff_pos);

# ----------------------
# --- used variables ---
# ----------------------
our ($opt_s, $opt_S, $opt_a, $opt_e, $opt_c, $opt_k, $opt_t, $opt_i, $opt_f);
our ($opt_g, $opt_d, $opt_D, $opt_h, $opt_I, $opt_L, $opt_n, $opt_v);
our ($opt_V, $opt_T, $opt_C, $opt_l, $opt_r, $debug);

my  ($s_exp_file, $s_act_file, $s_col_file);

my  $s_version = "2.0";    # Version von csvdiff
my  $s_use_key = 1;        # Vergleich mit (=1) oder ohne (=0) Key
my  $s_key_num = 1;        # Key numerisch (=1) oder String/Feldname (=0)
our $s_key_cols = 0;       # Welche Spalte enthaelt den "unique key", Zaehlung
                           #  beginnt bei 0
my  $s_fadeout_cols = "";  # Welche Spate(n) sollen nicht mit verglichen werden
my  $s_ratio_cols = "";    # Fuer welche Spate(n) sollen ein Verhaeltnis und
                           #  Differenz berechnet werden
our $s_num_cols = "";      # Welche Spalten sollen numerisch verglichen werden
our $s_num_cols_invers = 0;# Sollen die uebergebenen Spalten numerisch verglichen
                           #  werden (0), oder alle anderen Spalten ausser den in
                           #  $s_num_cols abgelegten Spalten (1).
our @a_num_cols = ();      # Aufteilen der in $s_num_cols uebergebene(n) Felder
                           #  in einzelne Werte
#   $s_joined_key = "";    # Alle Keys zu einem String zusammenfassen, um diese
                           #  dann als Schluessel der Hashs zu verwenden
our $s_separator = ',';    # Feldtrenner
our $s_d_separator = '.';  # Dezimaltrennzeichen
our $s_t_separator = "";   # Tausendertrennzeichen
my  $s_verbose = 0;        # csvdiff "etwas" geschwaetziger machen
my  $s_sort_data = 0;      # Dateien sortieren vor vergleich?
our $s_ignore = 0;         # Gross - Kleinschreibung beim Vergleich beachten(=0),
                           #  oder nicht (=1)
our $s_ignore_space = 0;   # Doppelte Leerzeichen ignorieren
my  %h_key_act;            # Hash der aktuellen Daten
                           #  Schluessel = Keyspalte aus
                           #  csv Datei, Wert = Zeilennummer
my  %h_key_exp;            # Hash der erwarteten Daten
                           #  Schluessel = Keyspalte aus csv Datei
                           #  Wert = Zeilennummer
our %h_key_all;            # Hash das alle vorkommenden Key's enthaelt
                           #  Wert = wie oft kommt der jeweilige Key in beiden
                           #  Dateien vor
#   @a_records=();         # Array das mehrfach verwendet wird, um einen
                           #  Record am Trenner in einzelne Felder aufzuspalten
#   @a_key_cols=();        # Array in das die uebergebenen Keys aufgespalten
                           #  werden
our @a_fadeout_cols=();    # Array in das die uebergebenen "Fadeout" Spalten
                           #   aufgespalten werden
my  @a_ratio_cols=();      # Array in das die uebergebenen "Ratio" Spalten
                           #   aufgespalten werden
my  @a_lines_act=();       # Array das alle Zeilen aus dem ACT-File aufnimmt
my  @a_lines_exp=();       # Array das alle Zeilen aus dem EXP-File aufnimmt
my  @a_lines_col=();       # Array das alle Zeilen aus dem COL-File aufnimmt
                           #  eigentlich unnoetig, da COL nur eine Zeile
                           #  enthalten darf, so sieht's aber "symetrischer" aus
my  @a_col_names=();       # Array das die Spaltennamen aufnimmt
my  $s_diff_counter=0;     # Variable die die Anzahl der unterschiedlichen Zeilen
                           #  zaehlt
our $coloured_output=0;    # Soll der output diff aehnlich sein, aber mit Farbe?
our $show_pos_length=15;   # min. Laenge einer Zeichenkette, ab der die Position
                           #  des ersten Unterschieds angezeigt wird
my  $s_no_diff_per_record=0;
                           # Der wievielte Unterschied wurde pro Zeile gefunden?
                           #  um Leerzeile einzufügen zwischen 2 Unterschieden
our $s_compare_message=""; # Hierin wird die Warnung aus cmp_data aufgenommen,
                           #  wenn ein Feld das numerisch verglichen werden soll
                           #  nicht numerische Zeichen enthaelt


# Einige temporaere Variablen, die z.B. als Counter verwendet werden
my  $s_line_counter;       # Zeilenzaehler, wird mehrfach verwendet
my  $s_field_counter;      # Feldzaehler, wird mehrfach verwendet
our ($s_trim, $s_zeile, $s_i, $s_key, $s_diff, $s_col);
#my (@a_trim_act, @a_trim_exp, $s_tmp_var, $s_zeile_join);
our (@a_tmp_records);



# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#                                  M A I N
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#getopts('hdsae') or die "Option not valid!\n";
getopts('hdDtiIvVgClf:s:S:a:e:c:k:L:n:T:r:') or die "One Option is not valid!\n";
# siehe http://www.interaktiv.net/shop-system-support/tutorials/perlgetopt.shtml
# siehe http://www.cis.uni-muenchen.de/~hbosk/perl2_ws03/einlesen/optionen_einlesen.html

if (defined $opt_h) {
  #"Online Help" ausgeben
  print "\nOnline Help\n";
  usage;
  $opt_v = 1;
  exit 1;
}

if (defined $opt_v) {
  #Version ausgeben
  print "\nThis is version $s_version of csvdiff\n\n";
  print "csvdiff homepage: http://csvdiff.sourceforge.net/\n";
  print "         contact: r-sch\@users.sourceforge.net\n\n";
  print "This software is licenced unter the terms of GPL version 2.1, feel free\n";
  print "to use, customise, ... it.\n";
  print "If you implement new features, it will be fine if you contact me.\n";
  print "If you think there are missing features, or a bug please contact me too.\n";
  print "In a later version there will be more help, sorry.\n";
  exit 1;
}

if (defined $opt_T) {
  #Tausendertrenner
  if (defined $opt_n) {
    if($opt_T eq "" or $opt_T eq "." or $opt_T eq ","){
      $s_t_separator = $opt_T;
      if(defined $opt_S) {
        if($opt_T eq $opt_S) {
          print STDERR "The decimalseparator and the 1000 separator you provided \"$opt_S\"\n";
          print STDERR "can not be equal! => exit 1\n";
          exit 1;
        }
      }
    }
    else {
      print STDERR "The 1000 separator \"$opt_T\" you provided is not valid!\n";
      print STDERR "Only , or . or empty (no separator) are valid!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    print STDERR "You provided a 1000 separator, but no column to be compared numerical!\n";
    print STDERR "=> exit 1\n";
    exit 1;
  }
}
else {
  $s_t_separator = "";
}

if (defined $opt_S) {
  #Dezimaltrenner
  if (defined $opt_n) {
    if($opt_S eq "." or $opt_S eq ","){
      $s_d_separator = $opt_S;
      if(defined $opt_T) {
        if($opt_T eq $opt_S) {
          print STDERR "The decimalseparator and the 1000 separator you provided \"$opt_S\"\n";
          print STDERR "can not be equal! => exit 1\n";
          exit 1;
        }
      }
    }
    else {
      print STDERR "The Decimalseparator \"$opt_S\" you provided is not valid!\n";
      print STDERR "Only , or . are valid Decimalseparators!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    print STDERR "You provided a decimalseparator, but no column to be compared numerical!\n";
    print STDERR "=> exit 1\n";
    exit 1;
  }
}
else {
  $s_d_separator = ".";
}

if (defined $opt_s) {
  #Spaltentrenner
  $s_separator = $opt_s;
}
else {
  $s_separator = ",";
}

if (defined $opt_k) {
  #Keyspalte(n)
  $s_key_cols = $opt_k;
  $s_use_key = 1;

  my $tmp_check_key_col = $s_key_cols;
  #pruefen ob ein nichtnumerisches Zeichen (Spaltentrenner) enthalten ist
  #Achtung: hier wird ausschliesslich "formal" geprueft, ob der richtige
  #Spaltentrenner verwendet wurde

  if ($tmp_check_key_col =~ /\D+/x) {
    #das auftreten des gewuenschten Spaltentrenners durch nichts ersetzen
    $tmp_check_key_col =~ s/\Q$s_separator\E//xg;
    #nochmals pruefen ob ein nichtnumerisches Zeichen enthalten ist, wenn jetzt
    #noch eins da ist wurde ein andererer Trenner verwendet
    if ($tmp_check_key_col =~ /\D+/x) {
      print STDERR "The Columseparator you provided for keycolumns is not the one you used with the \"-s\" option!\n";
      print STDERR "You provided \"$s_key_cols\" as key column(s) and \"$s_separator\" as separator!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    ; #es wurde ein Integer uebergeben, hier muss "nur" geprueft werden ob
      #diese Spalte existiert, das passiert weiter unten
  }
}
else {
  $s_key_cols = 0;
  $s_use_key  = 0;
}

if (defined $opt_f) {
  #Spalte(n) ausblenden
  $s_fadeout_cols = $opt_f;
  my $tmp_check_fadeout_col = $s_fadeout_cols;
  #pruefen ob ein nichtnumerisches Zeichen enthalten ist
  #Achtung: hier wird ausschliesslich "formal" geprueft, ob der richtige
  #Spaltentrenner verwendet wurde
  if ($tmp_check_fadeout_col =~ /\D+/x) {
    #das auftreten des gewünschten Spaltentrenners durch nichts ersetzen
    $tmp_check_fadeout_col =~ s/\Q$s_separator\E//xg;
    #nochmals prüfen ob ein nichtnumerisches Zeichen enthalten ist, wenn jetzt
    #noch eins da ist würde ein andererer Trenner verwendet werden
    if ($tmp_check_fadeout_col =~ /\D+/x) {
      print STDERR "The Columseparator you provided for \"fadeout\" columns is not the one you used with the \"-s\" option!\n";
      print STDERR "You provided \"$s_fadeout_cols\" as fadeout column(s) and \"$s_separator\" as separator!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    ; #es wurde ein Integer uebergeben, hier muss "nur" geprueft werden ob
      #diese Spalte existiert, das passiert weiter unten
  }
}

if (defined $opt_n) {
  #Spalte(n) numerisch vergleichen
  $s_num_cols = $opt_n;

  #Ist das erste Zeichen das übergeben wurde ein Minus?
  #Wenn ja sollen alle Spalten ausser der/den angegebenen numerisch verglichen
  #werden!
  $s_num_cols_invers = ($s_num_cols =~ tr /-//d); #Wertebereich: 0 und 1
  #$s_num_cols_invers = 1 if($s_num_cols =~ /-.*/);
  #$s_num_cols =~ tr /-//d;

  #Sonderbehandlung wenn alle Spalten numerisch verglichen werden sollen
  $s_num_cols = 0 if($s_num_cols eq "");

  my $tmp_check_num_cols = $s_num_cols;
  #pruefen ob ein nichtnumerisches Zeichen enthalten ist
  #Achtung: hier wird ausschliesslich "formal" geprueft, ob der richtige
  #Spaltentrenner verwendet wurde
  if ($tmp_check_num_cols =~ /\D+/x) {
    #das auftreten des gew nschten Spaltentrenners durch nichts ersetzen
    $tmp_check_num_cols =~ s/\Q$s_separator\E//xg;
    #nochmals pruefen ob ein nichtnumerisches Zeichen enthalten ist, wenn jetzt
    #noch eins da ist wurde ein andererer Trenner verwendet
    if ($tmp_check_num_cols =~ /\D+/x) {
      print STDERR "The Columseparator you provided for \"numeric comparision\" columns is not the one you used with the \"-s\" option!\n";
      print STDERR "You provided \"$s_num_cols\" as column(s) to be numeric compared and \"$s_separator\" as separator!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    ; #es wurde ein Integer uebergeben, hier muss "nur" geprueft werden ob
      #diese Spalte existiert, das passiert weiter unten
  }
  #dieses Array enthaelt die Felder die numerisch verglichen werden sollen
###  @a_num_cols = split(/\Q$s_separator\E/x, $s_num_cols);
  @a_num_cols = parse_line($s_separator, $s_num_cols);
}

if (defined $opt_r) {
  if(defined $opt_d) {
    print STDERR "\nWARNING: Option -r will be IGNORED, because you used option -d too\n";
    $opt_r = undef;
  }
  #Spalte(n) fuer die ein Verhaeltnis und Unterschied wird gerechnet
  $s_ratio_cols = $opt_r;
  my $tmp_check_ratio_col = $s_ratio_cols;
  #pruefen ob ein nichtnumerisches Zeichen enthalten ist
  #Achtung: hier wird ausschliesslich "formal" geprueft, ob der richtige
  #Spaltentrenner verwendet wurde
  if ($tmp_check_ratio_col =~ /\D+/x) {
    #das auftreten des gewuenschten Spaltentrenners durch nichts ersetzen
    $tmp_check_ratio_col =~ s/\Q$s_separator\E//xg;
    #nochmals pruefen ob ein nichtnumerisches Zeichen enthalten ist, wenn jetzt
    #noch eins da ist wurde ein andererer Trenner verwendet
    if ($tmp_check_ratio_col =~ /\D+/x) {
      print STDERR "The Columseparator you provided for \"ratio\" columns is not the one you used with the \"-s\" option!\n";
      print STDERR "You provided \"$s_ratio_cols\" as ratio column(s) and \"$s_separator\" as separator!\n";
      print STDERR "=> exit 1\n";
      exit 1;
    }
  }
  else {
    ; #es wurde ein Integer uebergeben, hier muss "nur" geprueft werden ob
      #diese Spalte existiert, das passiert weiter unten
  }
###  @a_ratio_cols = split(/\Q$s_separator\E/x, $s_ratio_cols);
  @a_ratio_cols = parse_line($s_separator, $s_ratio_cols);
}

if (defined $opt_g) {
  #Daten vor vergleich sortieren?
  $s_sort_data = $opt_g;
}
else {
  $s_sort_data = 0;
}

if (defined $opt_D) {
  #Debug Modus
  $debug = 1;
}
else {
  $debug = 0;
}

if (defined $opt_V) {
  #Etwas mehr Output erzeugen
  $s_verbose = 1;
}
else {
  $s_verbose = 0;
}

if (defined $opt_d) {
  #diff aehnlicher Output, auf Linux mit Farbe
  $coloured_output = 1;
}
else {
  $coloured_output = 0;
}

if (defined $opt_t) {
  #Vor- und Nachlaufender Leerzeichen ignorieren
  $s_trim = 1;
}
else {
  $s_trim = 0;
}

if (defined $opt_i) {
  #Gross-/Kleinschreibung ignorieren
  $s_ignore = 1;
}
else {
  $s_ignore = 0;
}

if (defined $opt_I) {
  #Doppelte Leerzeichen ignorieren
  $s_ignore_space = 1;
}
else {
  $s_ignore_space = 0;
}

if (defined $opt_L) {
  #Min. laenge ab der die Position des ersten Unterschieds angezeigt wird

  #pruefen ob ein nichtnumerisches Zeichen enthalten ist
  #siehe auch perldoc -q "whether a scalar is a number"
  if($opt_L =~ /\D+/x) {
    print STDERR "You provided \"$opt_L\" as min. length to show the first difference, this is not an integer (!).\n";
    print STDERR "=> exit 1\n";
    exit 1;
  }
  else {
    $show_pos_length = $opt_L;
  }
}
else {
  $show_pos_length = 15;
}

# Pruefen ob beide Optionen zur uebergabe der Spaltennamen verwendet wurden
if($opt_c && $opt_C) {
  print STDERR "You can just use \"-c $opt_c\" or \"-C\", but not together!\n";
  print STDERR "=> exit 1\n";
  exit 1;
}

# -----------------------------------------------------------------------------
# Ausgeben aller aktivierten Optionen, da Verbose Modus
# -----------------------------------------------------------------------------
if($s_verbose) {
  print "This is csvdiff version: $s_version\n";
  print "Folowing Options are activated:\n";
  print "  Actual Result                 : $opt_a\n";
  print "  Expected Result               : $opt_e\n";
  print "  Columnamesfile                : $opt_c\n" if($opt_c);
  if($s_separator eq " "){
    print "  Fieldseparator                : \"$s_separator\"\n";
  }
  else{
    print "  Fieldseparator                : $s_separator\n";
  }
  print "  Show complete Line            : yes\n" if($opt_l);
  print "  Fade-out column(s)            : $opt_f\n" if($opt_f);
  print "  Ratio column(s)               : $opt_r\n" if($opt_r);
  if($opt_n) {
    print "  Decimalseparator              : $s_d_separator\n";
    if($s_t_separator) {
      print "  1000 separator                : $s_t_separator\n";
    }
    else {
      print "  1000 separator                : None\n";
    }
  }
  if($s_num_cols_invers) {
    print "  NOT Numeric compared column(s): $s_num_cols\n" if($opt_n);
  }
  else {
    print "  Numeric compared column(s)    : $s_num_cols\n" if($opt_n);
  }
  print "  Indicate first difference pos.: $show_pos_length\n" if($show_pos_length > 0);
  print "  Print online Help             : yes\n" if($opt_h);
  print "  Debug mode                    : yes\n" if($opt_D);
  print "  Diff like coloured output     : yes\n" if($opt_d);
  print "  Trim space's                  : yes\n" if($opt_t);
  print "  Ignore case                   : yes\n" if($opt_i);
  print "  Ignore multiple space's       : yes\n" if($opt_I);
  print "  Grade/sort input data first   : yes\n" if($opt_g);
  print "  Keycolumn(s)                  : $opt_k\n" if($opt_k);
  print "\n" if(!$opt_k); #Leerzeile nur wenn nicht noch die Spaltennamen
                          #ausgegeben werden sollen
}

# -----------------------------------------------------------------------------
# Pruefen ob die Datei $s_act_file existiert und lesbar ist
# -----------------------------------------------------------------------------
$s_act_file = chk_input_file($opt_a, 1, "Actual Result File");
$s_exp_file = chk_input_file($opt_e, 1, "Expected Result File");
$s_col_file = chk_input_file($opt_c, 0, "Columnames File");

# -----------------------------------------------------------------------------
# Dateien in Arrays einlesen, hierbei werden Leerzeilen (Zeilen ohne ein
# Zeichen, Zeilen, bei denen alle Felder NULL enthalten sind damit nicht
# gemeint) ignoriert!
# -----------------------------------------------------------------------------
if($opt_f) {
  # Es sollen eine oder mehrere Spalten nicht zum Vergleichen berücksichtigt
  # werden, die uebergebenen Spalten am Trenner aufspalten
###  @a_fadeout_cols = split(/\Q$s_separator\E/x, $s_fadeout_cols);
  @a_fadeout_cols = parse_line($s_separator, $s_fadeout_cols);
}

my $s_empty_lines_act = 0;
my $s_first_line_act = "";
#Daten aus ACT Datei einlesen
($s_empty_lines_act, $s_first_line_act, @a_lines_act) = read_csv($s_act_file);

my $s_empty_lines_exp = 0;
my $s_first_line_exp = "";
#Daten aus EXP Datei einlesen
($s_empty_lines_exp, $s_first_line_exp, @a_lines_exp) = read_csv($s_exp_file);

#Wenn die Ergebnisdateien die Spaltenüberschriften in der ersten Zeile
#enthalten, stimmen sie überein? Wenn ja, die Zeile aufspalten und an
#das Spaltennamensarray zuweisen.
if($opt_C) {
  if($s_first_line_act ne $s_first_line_exp) {
    print STDERR "Column names in actual and expected result are not equal => exit 1\n";
    exit 1;
  }
  else {
    if($s_first_line_act eq "") {
      print STDERR "\nThe first line in your Files should contain the Columnames, but it is empty => exit 1\n";
      exit 1;
    }
###    @a_col_names = split(/\Q$s_separator\E/x, $s_first_line_act);
    @a_col_names = parse_line($s_separator, $s_first_line_act);
  }
}

#Warnungen ausgeben wenn Leerzeilen in den Dateien enthalten sind
if($s_empty_lines_act > 0 or $s_empty_lines_exp > 0){
  print STDERR "\nWARNING: Empty lines are IGNORED, they are not counted etc., take care of linenumbers!\n";
  if($s_empty_lines_act > 0){
    print STDERR "  Actual Result contains $s_empty_lines_act empty lines.\n";
  }
  if($s_empty_lines_exp > 0){
    print STDERR "  Expected Result contains $s_empty_lines_exp empty lines.\n";
  }
  print STDERR "\n";
}

if($s_col_file) {
  open( "FILE_COL", "<", "$s_col_file" ) or croak "File $s_col_file could not be opend: $!";
    @a_lines_col = <FILE_COL>; # komplette Datei in Array einlesen
  close( FILE_COL );
  if (scalar @a_lines_col > 1) {
    print STDERR "Error: Your columnnames file has more than one line => exit 1\n";
    exit 1;
  }
  chomp(@a_lines_col);
  #ganze Array Zeile am Trenner aufspalten
###  @a_col_names = split(/\Q$s_separator\E/x, $a_lines_col[0]);
  @a_col_names = parse_line($s_separator, $a_lines_col[0]);
}

chomp(@a_lines_act, @a_lines_exp);

#Hier noch die Namen der Keyspalten ausgeben
if($s_verbose) {
  if($opt_k) {
    my @a_tmp_key_cols = split(/\Q$s_separator\E/x, $opt_k);
    print "  Keycolumnname(s)              : ";
    my $s_erster_key = 0;
    foreach my $s_tmp_spalte (@a_tmp_key_cols) {
      print "                                  " if($s_erster_key);
      $s_erster_key++;
      printf ("%-4d = $a_col_names[$s_tmp_spalte-1]\n",$s_tmp_spalte); #Die -1 ist wieder zur Berichtigung der Feldname des ersten Feldes steht in Arrayelement 0!
    }
  print "\n";
  }
}


# -----------------------------------------------------------------------------
# Fuehrende und nachlaufende Leerzeichen entfernen, sowie doppelte Leerzeichen
# -----------------------------------------------------------------------------
#if($s_ignore_space or $s_trim) {
#  @a_lines_act = trim_delete_space(@a_lines_act);
#  @a_lines_exp = trim_delete_space(@a_lines_exp);
#}

# -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Vergleich mit Key
# -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
if($s_use_key) {
  if($s_sort_data) {
    print "\nTo use a Keycolumn and the sort/grade Option together makes no sense, sort will be ignored!\n\n";
    $s_sort_data = 0;
  }

  # ---------------------------------------------------------------------------
  # Hash mit den Daten aus der "actual" bzw. "expected" Datei fuellen
  # ---------------------------------------------------------------------------
  %h_key_act = create_hash("act", @a_lines_act);
  %h_key_exp = create_hash("exp", @a_lines_exp);

  # ---------------------------------------------------------------------------
  # Jetzt/hier geht das Vergleichen los
  # ---------------------------------------------------------------------------
  $s_line_counter = 0;
  $s_diff_counter = 0;

  # Der Vergleich erfolgt in der sortierten Reihenfolge der Keys!
  foreach my $s_key ( sort keys %h_key_all){
    $s_line_counter++; # Zeilennummer inkrementieren

    # -------------------------------------------------------------------------
    # Key ist nur in exp enthalten
    # -------------------------------------------------------------------------
    if((defined $h_key_exp{$s_key}) and not (defined $h_key_act{$s_key})) {
      if($s_diff_counter > 0) {
        # Trennzeile ausgeben
        print "\n", '-' x 79, "\n\n";
      }
      $s_diff_counter++;
      if($opt_l) {
        print "Record with Key: \"$s_key\" exists only in expected result $s_exp_file\n";
        printf (" Line %03d : > $a_lines_exp[$h_key_exp{$s_key}] <\n",$h_key_exp{$s_key}+1);
      }
      else {
        printf( "Record with Key: \"$s_key\" exists only in expected result $s_exp_file, line %d\n",$h_key_exp{$s_key}+1);
      }
    }

    # -------------------------------------------------------------------------
    # Key ist nur in act enthalten
    # -------------------------------------------------------------------------
    elsif((not defined $h_key_exp{$s_key}) and ( defined $h_key_act{$s_key})) {
      if($s_diff_counter > 0) {
        # Trennzeile ausgeben
        print "\n", '-' x 79, "\n\n";
      }
      $s_diff_counter++;
      if($opt_l) {
        print "Record with Key: \"$s_key\" exists only in actual result $s_act_file\n";
        printf (" Line %03d : > $a_lines_act[$h_key_act{$s_key}] <\n",$h_key_act{$s_key}+1);
      }
      else {
        printf ("Record with Key: \"$s_key\" exists only in actual result $s_act_file, line %d\n",$h_key_act{$s_key}+1);
      }
    }

    # -------------------------------------------------------------------------
    # Key ist in act und exp enthalten
    # -------------------------------------------------------------------------
    elsif(defined $h_key_exp{$s_key} && defined $h_key_act{$s_key}) {
      my $s_act_line = $h_key_act{$s_key};
      my $s_exp_line = $h_key_exp{$s_key};

      $s_diff = 0;
      my $s_record_header = "";
      my $s_record_diff_line = "";
      $s_diff = cmp_data($a_lines_act[$s_act_line], $a_lines_exp[$s_exp_line]);
      # -----------------------------------------------------------------------
      # Die einzelnen Records (als ganzes) sind unterschiedlich
      # -----------------------------------------------------------------------
      if($s_diff) {
        if($s_diff_counter > 0) {
          # Trennzeile ausgeben
          #print "\n", '-' x 79, "\n\n";
          $s_record_header .= "\n" . ('-' x 79) . "\n\n";
        }
        $s_diff_counter++;
#        print "Record with key \"$s_key\" is different:\n";
        $s_record_header .= "Record with key \"$s_key\" is different:\n";

        #Zeile in einzelne Felder aufspalten
###        my @a_fields_act = split (/\Q$s_separator\E/x, $a_lines_act[$s_act_line]);
###        my @a_fields_exp = split (/\Q$s_separator\E/x, $a_lines_exp[$s_exp_line]);
        my @a_fields_act = parse_line($s_separator, $a_lines_act[$s_act_line]);
        my @a_fields_exp = parse_line($s_separator, $a_lines_exp[$s_exp_line]);

        # ---------------------------------------------------------------------
        # Unterschiede werden diff-maessig angezeigt
        # ---------------------------------------------------------------------
        if($coloured_output) {
          my $tmp_act_line = sprintf ("Actual   line %03d > ",$s_act_line+1);
          my $tmp_exp_line = sprintf ("Expected line %03d > ",$s_exp_line+1);

          $s_field_counter = 0;
          my $b_record_header_already_printed = 0;
          #Lauf ueber alle Felder der Records, so sie existieren
          while(defined $a_fields_act[$s_field_counter] && defined $a_fields_exp[$s_field_counter]) {
            $s_diff = 0;
            $s_diff = cmp_data($a_fields_act[$s_field_counter], $a_fields_exp[$s_field_counter], $s_field_counter);
            #Das aktuelle Feld ist unterschiedlich
            if($s_diff) {
              if($b_record_header_already_printed == 0) {
                print "$s_record_header";
                $b_record_header_already_printed=1;
              }
              #Farbige Ausgabe ist derzeit nur unter Linux moeglich, ggf. auch
              #unter anderen Unix'en, aber nicht unter Windows
              if(lc($^O) =~ "win") {
                $tmp_act_line .= "$a_fields_act[$s_field_counter]$s_separator";
                $tmp_exp_line .= "$a_fields_exp[$s_field_counter]$s_separator";
              }
              else {
                $tmp_act_line .= "\033[31m$a_fields_act[$s_field_counter]\033[39m$s_separator";
                $tmp_exp_line .= "\033[32m$a_fields_exp[$s_field_counter]\033[39m$s_separator";
              }
            }
            #Das aktuelle Feld ist gleich
            else {
              $tmp_act_line .= "$a_fields_act[$s_field_counter]$s_separator";
              $tmp_exp_line .= "$a_fields_exp[$s_field_counter]$s_separator";
            }
            $s_field_counter++;
          }

          $tmp_act_line .= " <\n";
          $tmp_exp_line .= " <\n";
          $s_record_diff_line .= $tmp_act_line;
          $s_record_diff_line .= $tmp_exp_line;

          if($b_record_header_already_printed) {
            print $s_record_diff_line;

            #Ausgeben/Anzeigen an welcher Stelle der erste Unterschied ist
            call_diff_pos($a_lines_act[$s_act_line], $a_lines_exp[$s_exp_line]);
          }
        }

        # ---------------------------------------------------------------------
        # Unterschiede werden im csvdiff "Format" angezeigt
        # ---------------------------------------------------------------------
        else {
          if($opt_l) {
            #Erstmal wird die ganze Zeile ausgegeben
            $s_record_diff_line .= sprintf ("Actual   line %03d > $a_lines_act[$s_act_line] <\n",$s_act_line +1);
            $s_record_diff_line .= sprintf ("Expected line %03d > $a_lines_exp[$s_exp_line] <\n",$s_exp_line +1);
          }
          else {
            #Nur die Zeilennummern ausgeben die verglichen werden
            $s_record_diff_line .= sprintf ("Actual line %d compared with Expected line %d\n",$s_act_line +1,$s_exp_line +1);
          }

          $s_field_counter = 0;
          $s_no_diff_per_record = 0;
          my $b_record_header_already_printed = 0;
          #Lauf ueber alle Felder der Records, so sie existieren
          while(defined $a_fields_act[$s_field_counter] && defined $a_fields_exp[$s_field_counter]) {
            $s_diff = 0;
            $s_compare_message = "";
            $s_diff = cmp_data($a_fields_act[$s_field_counter], $a_fields_exp[$s_field_counter], $s_field_counter);
            #Das aktuelle Feld ist unterschiedlich
            if($s_diff) {
              $s_no_diff_per_record++;
              #Ausgabe erfolgt nur wenn der erste Unterschied pro Zeile gefunden wird
              if($b_record_header_already_printed == 0) {
                print "$s_record_header";
                print "$s_record_diff_line";
                $b_record_header_already_printed=1;
              }
              print STDERR "$s_compare_message";
              print "\n" if($s_no_diff_per_record > 1); #Leerzeile einfuegen
                #zwischen zwei Unterschieden
              printf (" Difference in field no.: %02d",$s_field_counter +1);
              if ($opt_c || $opt_C) {
                print " - field name: $a_col_names[$s_field_counter]";
              }
              print "\n";
              print "  Actual   > $a_fields_act[$s_field_counter] <\n";
              print "  Expected > $a_fields_exp[$s_field_counter] <\n";

              if($opt_r) {
                #Pruefen ob ein Verhaeltnis und Unterschied fuer das aktuelle Feld erstellt werden soll
                if(grep{ $s_field_counter == ($_-1)} @a_ratio_cols) {
                  if(looks_like_number($a_fields_exp[$s_field_counter]) and looks_like_number($a_fields_act[$s_field_counter])) {
                    my $s_tmp_ratio = $a_fields_exp[$s_field_counter] / $a_fields_act[$s_field_counter];
                    my $s_tmp_diff = $a_fields_exp[$s_field_counter] - $a_fields_act[$s_field_counter];
                    print "   Ratio      > $s_tmp_ratio <\n";
                    print "   Difference > $s_tmp_diff <\n";
                  }
                  else {
                    print "   Ratio and difference couldn't be calculated!\n";
                  }

                }
              }

              #Ausgeben/Anzeigen an welcher Stelle der erste Unterschied ist
              call_diff_pos($a_fields_act[$s_field_counter],$a_fields_exp[$s_field_counter]);
            }
            #Das aktuelle Feld ist gleich
            else {
              ; #Hier wird nichts ausgegeben
            }
            $s_field_counter++;
          }

        }
      }

      # -----------------------------------------------------------------------
      # Die einzelnen Records (als ganzes) sind gleich
      # -----------------------------------------------------------------------
      else {
        # Die einzelnen Records (als ganzes) sind gleich
        #if($s_diff_counter > 0) {
        #  # Trennzeile ausgeben
        #  print "\n", '-' x 79, "\n\n";
        #}
        #$s_diff_counter++;
        #print "Record with key \"$s_key\" is idetical\n";
      }
    }

    # -------------------------------------------------------------------------
    # Key ist weder in act noch exp enthalten
    # -------------------------------------------------------------------------
    else {
      # Das duerfte eigentlich nie eintreten, koennte ggf. weggelassen werden,
      # und das letzte "elsif" zum "else" gemacht werden
      print STDERR "ERROR at Key: $s_key occoured\n";
      exit 1;
    }
  }
}
# -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Vergleich ohne Key
# -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
else {

  #Kopie des Arrays erstellen um die org. Zeilennummer nach dem Sortieren
  #ermitteln zu koennen
  my @a_lines_act_org = @a_lines_act;
  my @a_lines_exp_org = @a_lines_exp;

  #Daten vor vergleich sortieren
  if($s_sort_data) {
    @a_lines_act = sort sort_before_compare @a_lines_act;
    @a_lines_exp = sort sort_before_compare @a_lines_exp;
  }

  #Lauf ueber alle Zeilen beider Dateien, solange sie definiert sind
  for( $s_line_counter = 0; defined $a_lines_act[$s_line_counter] && defined $a_lines_exp[$s_line_counter]; $s_line_counter++) {

    my $s_record_header = "";
    my $s_record_diff_line = "";
    $s_diff = cmp_data($a_lines_act[$s_line_counter],$a_lines_exp[$s_line_counter], $s_field_counter);
    #In dieser Zeile gibt's mindestens einen Unterschied
    if($s_diff) {

      if($s_diff_counter > 0) {
        #Trennzeile ausgeben
        #print "\n", '-' x 79, "\n\n";
        $s_record_header .= "\n" . ('-' x 79) . "\n\n";
      }
      $s_diff_counter++;

      #Zeilen in einzelne Felder aufteilen
###      my @a_act_fields = split (/\Q$s_separator\E/x,$a_lines_act[$s_line_counter]);
###      my @a_exp_fields = split (/\Q$s_separator\E/x,$a_lines_exp[$s_line_counter]);
      my @a_act_fields = parse_line($s_separator,$a_lines_act[$s_line_counter]);
      my @a_exp_fields = parse_line($s_separator,$a_lines_exp[$s_line_counter]);
      #Variablen die nach und nach aufgebaut werden um den Output zu erzeugen
      my $tmp_act_line = "";
      my $tmp_exp_line = "";
      #Variablen fuer die orginalen Zeilennummern (wenn sortiert wurde)
      my $s_act_line_no = 0;
      my $s_exp_line_no = 0;

      # -----------------------------------------------------------------------
      # Vergleich ohne Key-Spalte, Output: diff-maessig
      # -----------------------------------------------------------------------
      if($coloured_output) {
        if($s_sort_data) {
          #Orginale Zeilennummer ermitteln
          for my $ind (0 .. $#a_lines_act_org) {
            if($a_lines_act_org[$ind] eq $a_lines_act[$s_line_counter]) {
              $s_act_line_no = $ind;
              last;
            }
          }
          for my $ind (0 .. $#a_lines_exp_org) {
            if($a_lines_exp_org[$ind] eq $a_lines_exp[$s_line_counter]) {
              $s_exp_line_no = $ind;
              last;
            }
          }
          #my ($s_act_line_no) = grep{ $a_lines_act_org[$_] eq $a_lines_act[$s_line_counter] } 0..$#a_lines_act_org;
          #my ($s_exp_line_no) = grep{ $a_lines_exp_org[$_] eq $a_lines_exp[$s_line_counter] } 0..$#a_lines_exp_org;

          $tmp_act_line = sprintf ("Actual   line %03d > ",$s_act_line_no+1);
          $tmp_exp_line = sprintf ("Expected line %03d > ",$s_exp_line_no+1);

          #Wenn sortiert wurde mach die Ausgabe der Zeilennummer so keinen Sinn
          #hier muesste wenn die Orginal Zeilennummer ausgegeben werden
          #$tmp_act_line = sprintf ("Actual   line > ");
          #$tmp_exp_line = sprintf ("Expected line > ");
        }
        else {
          $tmp_act_line = sprintf ("Actual   line %03d > ",$s_line_counter+1);
          $tmp_exp_line = sprintf ("Expected line %03d > ",$s_line_counter+1);
        }

        $s_field_counter = 0;
        my $b_record_header_already_printed = 0;
        # Lauf ueber alle Felder der Records, so sie existieren
        while(defined $a_act_fields[$s_field_counter] && defined $a_exp_fields[$s_field_counter]) {
          $s_diff = cmp_data($a_act_fields[$s_field_counter], $a_exp_fields[$s_field_counter], $s_field_counter);
          # Das aktuelle Feld ist unterschiedlich
          if($s_diff) {
            if($b_record_header_already_printed == 0) {
              print "$s_record_header";
              $b_record_header_already_printed=1;
            }
            #Farbige Ausgabe ist derzeit nur unter Linux moeglich, ggf. auch
            #unter anderen Unix'en, aber nicht unter Windows
            if(lc($^O) =~ "win") {
              $tmp_act_line .= "$a_act_fields[$s_field_counter]$s_separator";
              $tmp_exp_line .= "$a_exp_fields[$s_field_counter]$s_separator";
            }
            else {
              $tmp_act_line .= "\033[31m$a_act_fields[$s_field_counter]\033[39m$s_separator";
              $tmp_exp_line .= "\033[32m$a_exp_fields[$s_field_counter]\033[39m$s_separator";
            }
          }
          else {
            $tmp_act_line .= "$a_act_fields[$s_field_counter]$s_separator";
            $tmp_exp_line .= "$a_exp_fields[$s_field_counter]$s_separator";
          }
          $s_field_counter++;
        }

        $tmp_act_line .= " <\n";
        $tmp_exp_line .= " <\n";
        $s_record_diff_line .= $tmp_act_line;
        $s_record_diff_line .= $tmp_exp_line;

        if($b_record_header_already_printed) {
          print $s_record_diff_line;

          #Ausgeben/Anzeigen an welcher Stelle der erste Unterschied ist
          call_diff_pos($a_lines_act[$s_line_counter], $a_lines_exp[$s_line_counter]);
        }
      }

      # -----------------------------------------------------------------------
      # Vergleich ohne Key-Spalte, Output: standard
      # -----------------------------------------------------------------------
      else {

        if($s_sort_data) {
          #Orginale Zeilennummer ermitteln
          for my $ind (0 .. $#a_lines_act_org) {
            if($a_lines_act_org[$ind] eq $a_lines_act[$s_line_counter]) {
              $s_act_line_no = $ind;
              last;
            }
          }
          for my $ind (0 .. $#a_lines_exp_org) {
            if($a_lines_exp_org[$ind] eq $a_lines_exp[$s_line_counter]) {
              $s_exp_line_no = $ind;
              last;
            }
          }
          #my ($s_act_line_no) = grep{ $a_lines_act_org[$_] eq $a_lines_act[$s_line_counter] } 0..$#a_lines_act_org;
          #my ($s_exp_line_no) = grep{ $a_lines_exp_org[$_] eq $a_lines_exp[$s_line_counter] } 0..$#a_lines_exp_org;

          if($opt_l) {
            #Erstmal wird die ganze Zeile ausgegeben
            $s_record_diff_line .= sprintf ("Actual   line %03d > $a_lines_act[$s_line_counter] <\n",$s_act_line_no + 1);
            $s_record_diff_line .= sprintf ("Expected line %03d > $a_lines_exp[$s_line_counter] <\n",$s_exp_line_no + 1);
          }
          else {
            #Nur die Zeilennummern ausgeben die verglichen werden
            $s_record_diff_line .= sprintf ("Actual line %d compared with Expected line %d\n",$s_act_line_no + 1,$s_exp_line_no + 1);
          }

          #Wenn sortiert wurde mach die Ausgabe der Zeilennummer so keinen Sinn
          #hier muesste wenn die Orginal Zeilennummer ausgegeben werden
          #$tmp_act_line = sprintf ("Actual   line > ");
          #$tmp_exp_line = sprintf ("Expected line > ");
        }
        else {
          if($opt_l) {
            #Erstmal wird die ganze Zeile ausgegeben
            $s_record_diff_line .= sprintf ("Actual   line %03d > $a_lines_act[$s_line_counter] <\n",$s_line_counter + 1);
            $s_record_diff_line .= sprintf ("Expected line %03d > $a_lines_exp[$s_line_counter] <\n",$s_line_counter + 1);
          }
          else {
            #Nur die Zeilennummern ausgeben die verglichen werden
            $s_record_diff_line .= sprintf ("Actual line %d compared with Expected line %d \n",$s_line_counter + 1,$s_line_counter + 1);
          }
        }
#        printf ("\nRecord no: %03d is different:\n", $s_line_counter+1);
#        print "Actual   line: $a_lines_act[$s_line_counter]\n";
#        print "Expected line: $a_lines_exp[$s_line_counter]\n";

        $s_field_counter = 0;
        $s_no_diff_per_record = 0;
        my $b_record_header_already_printed = 0;
        #Lauf ueber alle Felder der Records, so sie existieren
        while(defined $a_act_fields[$s_field_counter] && defined $a_exp_fields[$s_field_counter]){

          $s_compare_message = "";
          $s_diff = cmp_data($a_act_fields[$s_field_counter], $a_exp_fields[$s_field_counter], $s_field_counter);
          #Das aktuelle Feld ist unterschiedlich
          if($s_diff) {
            $s_no_diff_per_record++;
            if($b_record_header_already_printed == 0) {
              print "$s_record_header";
              print "$s_record_diff_line";
              $b_record_header_already_printed=1;
            }
            print STDERR "$s_compare_message";
            print "\n" if($s_no_diff_per_record > 1); #Leerzeile einfuegen
              #zwischen zwei Unterschieden
            printf (" Difference in field no.: %02d",$s_field_counter +1);
            if ($opt_c || $opt_C) {
              print " - field name: $a_col_names[$s_field_counter]";
            }
            print "\n";
            print "  Actual   > $a_act_fields[$s_field_counter] <\n";
            print "  Expected > $a_exp_fields[$s_field_counter] <\n";

            #Ausgeben/Anzeigen an welcher Stelle der erste Unterschied ist
            call_diff_pos($a_act_fields[$s_field_counter],$a_exp_fields[$s_field_counter]);
          }
          #Das aktuelle Feld ist gleich
          else {
            ; #Hier wird nichts ausgegeben
          }
          $s_field_counter++;
        }
      }
    }
    #In dieser Zeile gibt's keinen Unterschied
    else {
      ; #Hier wird nichts ausgegeben
      #print "Zeile: $s_line_counter gleich\n";
    }
  }

  #----------------------------------------------------------------------------
  # Ausgabe der Zeilen die nur in ACT enthalten sind
  #----------------------------------------------------------------------------
  if(scalar @a_lines_act > $s_line_counter) {
    print "\n", '=' x 79, "\n\n";
    print "Folowing lines exists only in actual result $s_act_file:\n";
    for my $s_i ($s_line_counter .. (scalar @a_lines_act -1)) {
      #print "\n", '-' x 79, "\n\n";
      if($opt_l) {
        printf (" Line %03d : > $a_lines_act[$s_i] <\n",$s_i +1);
      }
      else {
        printf (" Line %d\n",$s_i +1);
      }
    }
  }

  #----------------------------------------------------------------------------
  # Ausgabe der Zeilen die nur in EXP enthalten sind
  #----------------------------------------------------------------------------
  if(scalar @a_lines_exp > $s_line_counter) {
    print "\n", '=' x 79, "\n\n";
    print "Folowing lines exists only in expected result $s_exp_file:\n";
    for my $s_i ($s_line_counter .. (scalar @a_lines_exp -1)) {
      #print "\n", '-' x 79, "\n\n";
      if($opt_l) {
        printf (" Line %03d : > $a_lines_exp[$s_i] <\n",$s_i +1);
      }
      else {
        printf (" Line %d\n",$s_i +1);
      }
    }
  }
}

#------------------------------------------------------------------------------
# Hinweis, das Farbe derzeit nur unter Linux moeglich ist
#------------------------------------------------------------------------------
if(lc($^O) =~ "win" && $coloured_output == 1) {
  print "\nOutput Coloring is not available for Windows (yet)!\n";
}

#------------------------------------------------------------------------------
# Hinweis das die Anzeige des ersten Unterschieds bei "diff" Output und
#  numerischem Vergleich nicht wirklich clever ist.
#------------------------------------------------------------------------------
if($coloured_output && $s_num_cols && $show_pos_length > 0) {
  print "\nIt may bee not so good to show position of first differnce while using";
  print " diff like output, und numeric column comparision!\n";
}

# ./csvdiff17.pl -e ratio2.csv -a ratio1.csv -o out.txt -s ";" -C -k 1 -S "." -r 2 -n 2  

__END__
=head1 NAME

csvdiff - csvdiff is a perl script to diff/compare two csv files with the
possibility to select the separator. Differences will be shown like: "Column
XYZ in record 999" is different. After this, the actual and the expected result
for this column will be shown.

=head1 SYNOPSIS

  ./csvdiff.pl -e exp.csv -a act.csv -s ";" -c col_names.csv -k 2 -f "11;12"

=head1 DESCRIPTION

csvdiff is a perl script for comparing two files of comma-separated values with
each other. In contrast to standard diff, it will show the number of the record
where the difference occurs, the column number, and (if provided) the fieldname
which is different. The separator can be set to any value you want, not just a
comma. It also provides support for multiple column keys, the ability to ignore
case and trim leading/tailing spaces, and the ability to ignore selected columns
such as timestamps.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Roland Schmitz, E<lt>roli@E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010 by Roland Schmitz

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.1 or,
at your option, any later version of Perl 5 you may have available.


=cut
