package csvdiff_function;


# =============================================================================
# csvdiff_function.pm - compare two csv files, Copyrigth 2006 - today
#  Roland Schmitz
# =============================================================================
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation version 2.1 of the License.
#
# This program is distributed because i hope that it will be usefull, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License version
# 2.1 for more details.
# =============================================================================
# csvdiff is a Perl script to diff/compare two csv files with the possibility
# to select the separator. Differences will be shown like: "Column XYZ in
# record 999" is different. After this, the actual and the expected result for
# this column will be shown.
# =============================================================================
# Contact: r-sch AT users DOT sourceforge DOT net
# =============================================================================
# Version 2.0  - created: 2010-11-01 by Roland Schmitz, did some Perltidy
#                                       and Perlcritic + feedback from
#                                       Jeremy Courter: use quotes within data
#                                       Thomas Fahle: reduce complexity
#                changed: YYYY-MM-DD by xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
#            last change: YYYY-MM-DD by xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# =============================================================================

use 5.010001;
use strict;
use warnings;

use Carp;
use Scalar::Util qw[ looks_like_number ];

use Data::Dumper;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use csvdiff_function ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
	
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(usage buggy chk_input_file read_csv parse_line trim
  trim_delete_space create_hash sort_before_compare cmp_data
  diff_pos call_diff_pos
);

our $VERSION = '0.01';


# Preloaded methods go here.

# === Funktion ================================================================
# Name:           usage
# Beschreibung:   Ausgabe der "Onlinehilfe"
# Parameter:      -
# Rueckgabe Wert: -
# =============================================================================
sub usage {
#  print "@_\n";
  print "\nUsage:\n";
  print "======\n";
  print "  Parameters:\n";
  print "    -e <File1>             Expected Result\n";
  print "    -a <File2>             Actual Result\n\n";
  print "All folowing Parameters ar optional!\n";
  print "    -c <File3>             Columnames in csv format (in only one Line!)\n";
  print "    -C                     Columnames in first line of both result files\n";
  print "    -s <Separator>         Fieldseparator (default=,)\n";
  print "    -S <Decimalseparator>  Decimalseparator (default=.)\n";
  print "                            it can be . or ,\n";
  print "    -T <1000 Separator>    thousands separator (default=empty)\n";
  print "                            it can be empty (no separator) or . or ,\n";
  print "    -k <KeyCol(s)>         Keycolumn(s)\n";
  print "    -f <Col(s)>            Fade-out column(s)\n";
  print "    -n <Col(s)>            This column(s) should be numerical compared,\n";
  print "                            if the columnlist starts with - all columns EXPECT theese\n";
  print "                            are numerical compared, e.g. \"-2;3;5\"!\n";
  print "    -N                     All numeric columns are compared numerical, the rest will\n";
  print "                            be compared caracter by caracter. -n will be ignored if -N is set!\n";
  print "    -r <Col(s)>            A ratio and difference of these columns will be computed\n";
  print "                            NOT with option -d\n";
  print "                            Ratio = Expected/Actual   Difference = Expected-Actual\n";
  print "    -L <Length>            Min. length to show exact position of first difference,\n";
  print "                            optional (default=15)\n";
  print "    -h                     print online help\n";
  print "    -D                     Debug mode\n";
  print "    -d                     coloured Output like diff, ONLY for Linux\n";
  print "    -t                     Trim space's\n";
  print "    -i                     Ignore-case\n";
  print "    -I                     Ignore multiple space's\n";
  print "    -g                     Grade/sort data before comparision, this has only\n";
  print "                            effect when there are no key columns\n";
  print "    -v                     Print csvdiff version and quit\n";
  print "    -V                     Verbose mode\n";
  print "    -l                     Show complete lines which are compared, otherwise show\n";
  print "                            only linenumbers\n";
  print "\n";
  print "If your key consists of multiple fields/columns use for example\n";
  print "  -k \"1;4;2\", column one four and two are the unique key.\n";
  print "\n";
  print "Multiple keys has to be separated by the fieldseparator defind with Option \"-s\".\n";
  print "\n";
  print "If a column is min. $main::show_pos_length (default) characters long, the exact position of the\n";
  print "  first difference will be indicated by a ^ symbol, you can specify this by the\n";
  print "  \"-L\" option. If you don't want to use this feature, use \"-L 0\"!\n";
  print "\n";
  print "If TAB is your column serarator use -s \"\\t\" as separator.\n";
  print "Caution: empty lines/fields are displayed like: \">  <\" (note the two blanks)!\n";
  print "\nColumn and record count starts with 1!\n";
  print "\n";
  print "\nExample: perl csvdiff.pl -e exp.csv -a act.csv -s \";\" -c col_names.csv -k 2 -f \"11;12\"\n";

  return 1;
}
# === Ende von Funktion usage =================================================

# === Funktion ================================================================
# Name:           buggy
# Beschreibung:   Ausgabe von Debuginformationen auf STDERR
# Beispiel:       &buggy(__PACKAGE__, __LINE__, __FILE__, "Kommentar");
# Parameter:      Paketname
#                 Zeilennummer
#                 Dateiname
#                 zusaetzlicher Kommentar
# Rueckgabe Wert: -
# =============================================================================
sub buggy {
  my @input = @_;
  if(!$main::debug){
    return 1;
  }
  my $index=0;
  print STDERR "\n-------------------------\n";
  foreach(@input){
    $index++;
    if(!$_)
      { $_ = "" }
    if($index==1)
      { print STDERR "Package: " . $_ ."\n" }
    if($index==2)
      { print STDERR "Line:    " . $_ ."\n" }
    if($index==3)
      { print STDERR "File:    " . $_  ."\n" }
    if($index==4)
      { print STDERR "Comment: " . $_  ."\n" }
  }
  print "-------------------------\n";

  return 1;
}
# === Ende von Funktion buggy =================================================

# === Funktion ================================================================
# Name:           chk_input_file
# Beschreibung:   Pruefen ob die angegebene Datei existiert und lesbar ist
# Beispiel:       $s_act_file = chk_input_file($file, $required, $comment);
# Parameter:      $file     = Datei die eingelesen werden soll
#                 $required = Muss diese Datei existieren, wenn ja dann 1,
#                             wenn nein dann 0
#                 $comment  = Kommentar fuer die Ausgabe
#                             z.B. "Actual Result File"
# Rueckgabe Wert: $checked_file = Hier wird der Dateiname nach der Pruefung
#                                 fuer die weitere Verwendung abgelegt
# =============================================================================
sub chk_input_file {
  my ($tmp_file, $tmp_required, $tmp_comment) = @_;
  my $checked_file = "";

  #Datei wird benötigt, aber kein Name wurde übergeben
  if ( $tmp_required and not defined $tmp_file) {
    warn "Error:\n";
    warn "  $tmp_comment is needed! ->Exit\n";
    usage;
    exit 1;
  }

  #Datei wird NICHT benötigt und kein Name wurde übergeben
  if ( not defined $tmp_file) {
    return $checked_file;
  }

  #Datei wird benötigt, ist vorhanden und lesbar
  if ( -f $tmp_file and -r  $tmp_file ) {
    $checked_file = $tmp_file;
    return $checked_file;
  }

  # Nur im Fehlerfall kommen wir hin

  #Datei wird benötigt, ist aber nicht vorhanden
  if ( not -f $tmp_file ) {
    warn "  $tmp_comment doesn't exist";
    exit 1;
  }

  #Datei wird benötigt, ist vorhanden aber nicht lesbar
  if ( not -r $tmp_file ) {
    warn "  $tmp_comment is unreadable";
    exit 1;
  }

}

# === Ende von Funktion chk_input_file ========================================

# === Funktion ================================================================
# Name:           read_csv
# Beschreibung:   CSV Datei in Array einlesen, und Anzahl der Leerzeilen
#                 ermitteln, ggf. auch Felder ausblenden
# Beispiel:       ($leerzeilen, @datenzeilen) = read_csv($csv_datei);
# Parameter:      $csv_datei   = Name der einzulesenden CSV Datei
# Rueckgabe Wert: $leerzeilen  = Anzahl der Leerzeilen in der Datei
#                 $erste_zeile = Wenn opt_C gesetzt ist, enth��?t die erste
#                                Zeile die Spaltennamen
#                 @datenzeilen = Alle Zeilen die nicht leer sind
# =============================================================================
sub read_csv {
  my $tmp_filename = shift;
  #&buggy(__PACKAGE__, __LINE__, __FILE__, "Subroutine \"read_csv\" Anfang") if($debug);

  #Rueckgabewerte (-variablen) definieren
  my $s_tmp_empty_lines = 0; #wieviele Leerzeilen enthaelt die Datei
  my $s_tmp_first_line = ""; #wird fuer Spaltennamen verwendet, wenn vorhanden
  my @a_tmp_lines = ();      #hierin werden die Daten an das Hauptprogramm
                             #zurueckgegeben

  my $s_tmp_line_counter = 0;

  open( "CSV_FILE", "<", "$tmp_filename" ) or croak "File $tmp_filename could not be opened: $!";
    my $re_compiled = qr/ *(\Q$main::s_separator\E) */; #Vorkomplierter regulaerer
                                                  #Ausdruck, macht's schneller.
                                                  #siehe: perldoc perlop
    while (<CSV_FILE>) {
      $s_tmp_line_counter++;
      chomp $_;

      #Wenn die erste Zeile die Spaltennamen enthaelt, wird die komplette Zeile
      #der Rückgabevariablen $s_tmp_first_line zugewiesen
      if($s_tmp_line_counter==1 && $main::opt_C) {
        $s_tmp_first_line = $_;
        next;
      }
      #Wenn die Zeile leer ist, wird sie gezaehlt, und die Verarbeitung geht mit
      #der naechsten Zeile weiter
      if(!$_) {
        $s_tmp_empty_lines++;
        next;
      }
  #    next unless $_; #Leerzeilen werden ignoriert

      #Trim, Vor- und Nachlaufende Leerzeichen entfernen
      if($main::opt_t) {
        #Leerzeichen um den Feldtrenner herum loeschen
        #s/ *(\Q$s_separator\E) */$1/g;
        s/$re_compiled/$1/g;

        #Leerzeichen vor dem ersten und nach dem letzten Feld loeschen
        #s/^ *(.*?) *$/$1/;
        s/^ +//;
        s/ +$//;

        #Wenn ganz allgemein Whitespace's geloescht werden sollen sind folgende
        #zwei Zeilen zu aktivieren
        #s/\s*(\Q$s_separator\E)\s*/$1/g;
        #s/^\s*(.*?)\s*$/$1/;
      }

      #Doppelte Leerzeichen zu einem machen
      if($main::opt_I) {
        s/ +/ /g;
      }

      # Gross - Kleinschreibung ignorieren
      if($main::opt_i) {
        $_ = lc($_);
      }

      #Spalte(n) ausblenden
      if($main::opt_f) {
        #Zeile am Trenner aufspalten
###        @a_tmp_records = split(/\Q$s_separator\E/x, $_);
        my @a_tmp_records = parse_line($main::s_separator, $_);
        foreach my $s_col (@main::a_fadeout_cols) {
          my $s_tmp_col = $s_col-1;
          # Pruefen ob die angegebene Spalte existiert
          if(exists $a_tmp_records[$s_tmp_col]) {
            # Spalte existiert, Inhalt durch festen String ersetzen
            $a_tmp_records[$s_tmp_col] = "Fade out for comparing";
          }
          else {
            #print STDERR "Error: The fadeout column: $s_col did not exist => exit 1\n";
            #exit 1;
            print STDERR "\nWARNING: The fadeout column: $s_col did not exist in line $s_tmp_line_counter in $tmp_filename!\n";
          }
        }
        #print Dumper(@a_tmp_records);
        #Array wieder zum String machen
        my $s_tmp_zeile_join = join("$main::s_separator", @a_tmp_records);
        #aktuelle Zeile ans Ende des Arrays haengen
        push(@a_tmp_lines, $s_tmp_zeile_join);
      }
      #Nichts wird ausgeblendet
      else {
        #aktuelle Zeile einfach ans Ende des Arrays haengen
        push(@a_tmp_lines, $_);
      }
    }
  close( CSV_FILE );
  #print Dumper(@a_tmp_lines);

  return ($s_tmp_empty_lines, $s_tmp_first_line, @a_tmp_lines);
}
# === Ende von Funktion read_csv ==============================================

# === Funktion ================================================================
# Name:           parse_line
# Beschreibung:   Parses a line of delimited text.  If the Text::CSV module is 
#                 present, it will be used.  Otherwise, split() is used.
# Beispiel:       @a_fields = parse_line("Ein,Zwei,Drei",",");
# Parameter:      delimiter character
#                 line of delimited text
# Rueckgabe Wert: array of parsed data
# =============================================================================
sub parse_line {
        my @tmp_arr = @_;
	my $s_delimiter=shift(@tmp_arr);
	my $s_data=shift(@tmp_arr);
	my @a_data;
	
	#standardize line terminator
	#$s_data=~s/\r?\n?$/$\//;
	
	eval "use Text::CSV;"; ## no critic qw(BuiltinFunctions::ProhibitStringyEval)
	if ($@) {
		@a_data = split(/\Q$s_delimiter\E/x, $s_data);
	} else {
		my $csv = Text::CSV->new({sep_char => $s_delimiter});
		$csv->parse($s_data);
		my $tmp_rc = $csv->parse($s_data);
		die "A Error occured while parsing input file: $tmp_rc !\n" if(! $tmp_rc);

		@a_data = $csv->fields();
	}
	return @a_data;
}
# === Ende von Funktion parse_line ============================================

# === Funktion ================================================================
# Name:           create_hash
# Beschreibung:   Hash aus den in das jeweilige Array (@a_lines_act oder
#                 @a_lines_exp) eingelesenen Daten erzeugen, in dem dem Key die
#                 jeweilige Zeilennummer zugeordnet wird
# Beispiel:       %h_key = create_hash($file, @a_daten);
# Parameter:      $file    = "act" oder "exp", wird fuer die Fehlermeldung
#                            benoetigt
#                 @a_daten = Array, das die Daten (Records) aus der
#                            eingelesenen Datei enthaelt
# Rueckgabe Wert: %h_key = Hash der alle Key's (mittels der golbalen Variable
#                          $s_key_cols ermittelt) als Schluessel enthaelt, und
#                          die Zeilennummer als Wert
# =============================================================================
sub create_hash {

  my ($s_file, @a_daten) = @_;

  my $s_line_counter = 0;
  my @a_records = (); # Array leer machen
  my @a_key_cols = ();
  my %h_tmp_key_data;

  #Hash erzeugen in dem dem Key die Zeilennummer zugeordnet wird
  foreach my $s_zeile (@a_daten) {

    #ganze Array Zeile am Trenner aufspalten
###    @a_records  = split(/\Q$s_separator\E/x, $s_zeile);
    @a_records  = parse_line($main::s_separator, $s_zeile);

    #die uebergebenen Keys (Keyspalten) am Trenner aufspalten
###    @a_key_cols = split(/\Q$s_separator\E/x, $s_key_cols);
    @a_key_cols = parse_line($main::s_separator, $main::s_key_cols);

    #In dieser Variable wird der Unique Key zusammen gesetzt, wenn er aus
    #mehreren Spalten besteht
    my $s_joined_key = "";
    #Diese Variable wird auf 1 gesetzt wenn der Schluessel leer ist, dann wird
    #eine Warnung ausgegeben
    my $s_empty_key  = 0;

    my $s_tmp_line_no = 0;
    my $s_tmp_line_no_already_found = 0;

    #--------------------------------------------------------------------------
    #Schluessel fuer Hash aufbauen
    #--------------------------------------------------------------------------
    #Die -1 in den folgenden Zeilen ist noetig, da das erste Feld im Array die
    #Nr. 0 ist, uebergeben wird aber eine 1, was auch so beabsichtigt ist!

    #Der Key besteht nur aus einem Feld
    if(scalar @a_key_cols == 1) {
      #Pruefen ob der angegebene Key existiert
      if(exists $a_records[($a_key_cols[0]-1)]) {
        if(defined $a_records[($a_key_cols[0]-1)]) {
          #Key existiert und ist definiert
          if($a_records[($a_key_cols[0]-1)] eq "") {
            $s_empty_key = 1;
          }
        }
        else {
          $s_tmp_line_no = $s_line_counter + 1;
          print STDERR "Error: The keycolumn: $a_key_cols[0] contains an undefined value in $s_file file  => exit 1\n";
          print STDERR "Complete Line $s_tmp_line_no in $s_file file: > $s_zeile <\n\n" if($main::opt_l);
          exit 1;
        }
      }
      else {
        $s_tmp_line_no = $s_line_counter + 1;
        print STDERR "Error: The keycolumn: $a_key_cols[0] did not exist in $s_file file  => exit 1\n";
        print STDERR "Complete Line $s_tmp_line_no in $s_file file: > $s_zeile <\n\n" if($main::opt_l);
        exit 1;
      }
      $s_joined_key = $a_records[($a_key_cols[0]-1)]
    }

    #Der Key besteht aus mehreren Feldern
    else {
      my $s_tmp_counter = 0;
      foreach (@a_key_cols) {
        #Pruefen ob der angegebene Key existiert
        if(exists $a_records[($_-1)]) {
          if(defined $a_records[($_-1)]) {
            #Key existiert und ist definiert
            if($a_records[($_-1)] eq "") {
              #$s_empty_key = 1; #Bei zusammengesetzten Keys kann ein einzelnes Feld durchaus leer sein
            }
          }
          else {
            print STDERR "Error: The keycolumn: $_ contains an undefined value in $s_file file => exit 1\n";
            print STDERR "Complete Line in $s_file file: > $s_zeile <\n\n" if($main::opt_l);
            exit 1;
          }
        }
        else {
          print STDERR "Error: The keycolumn: $_ did not exist in $s_file file => exit 1\n";
          print STDERR "Complete Line in $s_file file: > $s_zeile <\n\n" if($main::opt_l);
          exit 1;
        }

        #Die einzelnen Keys mit dem Spaltentrenner zu einer Variable verbinden
        if($s_tmp_counter == 0) {
          #Sonderbehandlung fuer ersten Key, damit der Trenner nur zwischen den
          #  Keys steht, und nicht vor dem ersten
          $s_joined_key = $a_records[($_-1)];
        }
        else {
          $s_joined_key = $s_joined_key . $main::s_separator . $a_records[($_-1)];
        }

        $s_tmp_counter++;
      }
      my $s_is_key_empty;
      ($s_is_key_empty = $s_joined_key ) =~ s/\Q$main::s_separator\E//xg;
      #Bei zusammengesetzten Keys kann ein einzelnes Feld durchaus leer sein
      if($s_is_key_empty eq "") {
        $s_empty_key = 1;
      }
      else {
        $s_empty_key = 0;
      }
    }

    #Der Schluessel der eingefuegt werden soll existiert bereits im Hash
    if(defined $h_tmp_key_data{$s_joined_key}) {
      $s_tmp_line_no = $s_line_counter + 1;
      $s_tmp_line_no_already_found = $h_tmp_key_data{$s_joined_key} + 1;
      print STDERR "ERROR: Key \"$s_joined_key\" is not unique in $s_file file => exit 1!\n";
      print STDERR "This Key occurs in line $s_tmp_line_no_already_found and in line $s_tmp_line_no.\n";
      print STDERR "Complete Line $s_tmp_line_no_already_found: > $a_daten[$h_tmp_key_data{$s_joined_key}] <\n" if($main::opt_l);
      print STDERR "Complete Line $s_tmp_line_no: > $s_zeile <\n\n" if($main::opt_l);
      exit 1;
    }
    #Der Schluessel ist neu und wird eingefuegt
    else {
      $h_tmp_key_data{$s_joined_key} = $s_line_counter; #Hash belegen
      $s_line_counter++; # Zeilennummer inkrementieren
      warn "Do you think a key which contains NULL (is empty) is very clever?\n" if($s_empty_key);
    }
  }

  #Hash der alle Keys (aus beiden Dateien) enthaelt, mit Anzahl des vorkommens
  #als Wert. Ist der Wert 2, existiert der Key in beiden Dateien, ist er 1 nur
  #in einer der Dateien.
  #Werte ungleich 1 oder 2 duerfen nicht auftreten, da ein 0 maliges Vorkommen
  #nicht moeglich ist, und sollte der Wert groesser 2 werden, wuerde der Key in
  #mindestens einer der beiden Dateien mehrfach vorkommen, was weiter oben
  #bereits gefiltert wird. Das koennte heissen, das ggf. eine Spalte des Keys
  #vergessen wurde, wenn er aus mehreren besteht.
  foreach my $s_key ( sort keys %h_tmp_key_data){
    if(defined $main::h_key_all{$s_key}) {
      $main::h_key_all{$s_key} += 1;
      #zusaetzlicher Check, sollte eigentlich nie greifen!
      if($main::h_key_all{$s_key} > 2){
        print STDERR "ERROR: Key \"$s_key\" is not unique in $s_file file => exit 1!\n";
        print STDERR "Complete Line in $s_file file: $main::s_zeile\n\n" if($main::opt_l);
        exit 1;
      }
    }
    else {
      $main::h_key_all{$s_key}=1;
    }
  }

  return %h_tmp_key_data;
}
# === Ende von Funktion create_hash ===========================================

# === Funktion ================================================================
# Name:           trim
# Beschreibung:   Fuehrende nund nachlaufende Leerzeichen aus Strings oder
#                 Arrays entfernen
# Beispiel:       $string = trim($string);
#                 @many   = trim(@many);
# Parameter:      Zeichenkette oder Array die/das "getrimmt" werden soll
# Rueckgabe Wert: Zeichenkette oder Array je nach Aufruf, ohne fuehrende/
#                 nachlaufende Leerzeichen
# =============================================================================
sub trim {
  my @out = @_;
  &buggy(__PACKAGE__, __LINE__, __FILE__, "Subroutine \"trim\" Anfang") if($main::debug);
  for (@out) {
    #s/^\s*(.*?)\s*$/$1/g; #schlechte Loesung, dauert ca. 3 mal so lange!
    #s/^\s+//;
    #s/\s+$//;
    s/^ +//;
    s/ +$//;
  }
  return wantarray ? @out : $out[0];
}
# === Ende von Funktion trim ==================================================

# === Funktion ================================================================
# Name:           trim_delete_space
# Beschreibung:   Hier werden je nach angegebener Option fuehrende und nach-
#                 laufende Leerzeichen, sowie doppelte/mehrfache leerzeichen
#                 innerhalb der Felder entfernt, bzw. duch ein Leerzeichen
#                 ersetzt
# Beispiel:       @a_trimed_array = trim_delete_space(@a_array2trim);
# Parameter:      @a_array2trim   = Array, das die Daten (Records) enthaelt die
#                                   "getrimmt" werden sollen
# Rueckgabe Wert: @a_trimed_array = um "unnoetige" Leerzeichen bereinigtes
#                                   Array
# =============================================================================
sub trim_delete_space {
  my @tmp_array  = @_;
  my @a_records  = ();
  my @a_tmp_trim = ();

  foreach my $s_zeile (@tmp_array) {

    #Ein oder mehrmaliges Leerzeichen durch ein Leerzeichen ersetzen
    if($main::s_ignore_space) {
      $s_zeile =~ s/ +/ /g;
    }

    #Fuehrende und nachlaufende Leerzeichen abschneiden
    if($main::s_trim) {
      $s_zeile =~ s/^\s*(.*?)/$1/xg; #Leerzeichen vor dem ersten Feld
      $s_zeile =~ s/\s*(\Q$main::s_separator\E)\s*/$1/xg; #Leerzeichen um den Feldtrenner herum ersetzen
      $s_zeile =~ s/(.*?)\s*$/$1/xg; #Leerzeichen nach dem letzen Feld
    }

    #Bearbeitete Zeilen/Records wieder in ein Array schreiben
    push(@a_tmp_trim, $s_zeile);
  }

  return @a_tmp_trim;
}
# === Ende von Funktion trim_delete_space =====================================

# === Funktion ================================================================
# Name:           sort_before_compare
# Beschreibung:   Daten fuer das sortieren (mit sort) vergleichen
# Beispiel:
# Parameter:
# Rueckgabe Wert: Die beiden uebergebenen Variablen werden sortiert
#                 zurueckgegeben
# =============================================================================
sub sort_before_compare {
    return $a <=> $b if(looks_like_number($a) and looks_like_number($b));
    return lc $a cmp lc $b;
}
# === Ende von Funktion sort_before_compare ===================================

# === Funktion ================================================================
# Name:           cmp_data
# Beschreibung:   Vergleichen von zwei Feldern
# Beispiel:       $s_diff = cmp_data($rec1, $rec2, $akt_feld);
# Parameter:      $rec1 = Zeichenkette die den Datenrecord aus der ersten
#                         Datei enthaelt
#                 $rec2 = Zeichenkette die den Datenrecord aus der zweiten
#                         Datei enthaelt
#                 $akt_feld = Aktuelle Feldnummer, wird benoetigt um pruefen zu
#                             koennen, ob numerisch verglichen werden soll
# Rueckgabe Wert: 0 wenn's gleich ist, 1 wenn die Felder unterschiedlich sind
# =============================================================================
sub cmp_data {
  my ($s_tmp_data1, $s_tmp_data2, $s_tmp_field_no) = @_;
  &buggy(__PACKAGE__, __LINE__, __FILE__, "Subroutine \"cmp_data\" \ns_tmp_data1=$s_tmp_data1 \ns_tmp_data2=$s_tmp_data2") if($main::debug);
  my $s_tmp_cmp_res = undef;

  #Verschoben nach read_csv
  #if($s_ignore) {
  #  # Gross - Kleinschreibung ignorieren
  #  $s_tmp_data1 = lc($s_tmp_data1);
  #  $s_tmp_data2 = lc($s_tmp_data2);
  #}

  #Erstmal wird grundsaetzlich ein String Vergleich gemacht!
  #siehe prelre:
    #\A = Match only at beginning of string
    #\Z = Match only at Ende von string, or before newline at the end
    #\z = Match only at Ende von string
    #\E = end case modification (think vi)
    #\Q = quote (disable) pattern metacharacters till \E
  if($s_tmp_data1 =~ /\A\Q$s_tmp_data2\E\z/x) {
    $s_tmp_cmp_res = 0;
  }
  else {
    $s_tmp_cmp_res = 1;
  }

  #hier wird der komplette Record miteinander verglichen
  if(($s_tmp_data1 =~ /\Q$main::s_separator\E/x) or ($s_tmp_data2 =~ /\Q$main::s_separator\E/x)) {
    #der komplette Record wird ausschliesslich per String Vergleich verglichen
    #der String Vergleich wurde bereits durchgefuehrt!
    ;
  }

  #hier werden einzelne Felder miteinander verglichen
  else {
    #ACHTUNG, $s_tmp_cmp_res wurde auf jeden Fall schon durch den String
    #Vergleich mit einem Ergebnis belegt!!!

    #hier wird mal wieder 1 addiert, da in csvdiff die erste Spalte, Zeile, ...
    #die 1 ist und nicht die 0!
    $s_tmp_field_no += 1;

    #gibt es ueberhaupt Spalten die numerisch verglichen werden sollen?
    #wenn nein, ok dann wurde der Stringvergleich bereits gemacht
    if($main::s_num_cols){

      my $s_cmp_mode = "";
      #Alle Spalten sollen numerisch verglichen werden AUSSER die in
      #@a_num_cols enthaltenen
      if($main::s_num_cols_invers) {
        #pr��?en ob die aktuelle Spalte NICHT im Array der Spalten ist die
        #numerisch verglichen werden sollen
        if(!grep{ $main::a_num_cols[$_] eq $s_tmp_field_no } 0..$#main::a_num_cols) {
          $s_cmp_mode = "numeric";
        }
        else {
          $s_cmp_mode = "text";
        }
      }
      #Die in @a_num_cols enthaltenen Spalten sollen numerisch verglichen
      #werden
      else {
        #pr��?en ob die aktuelle Spalte im Array der Spalten ist die numerisch
        #verglichen werden sollen
        if(grep{ $main::a_num_cols[$_] eq $s_tmp_field_no } 0..$#main::a_num_cols) {
          $s_cmp_mode = "numeric";
        }
        else {
          $s_cmp_mode = "text";
        }
      }

      #Ueberpruefen ob die aktuelle Spalte numerisch verglichen werden soll
      if( $s_cmp_mode eq "numeric" ) {

        my $regex1 = '[0-9' . quotemeta($main::s_t_separator) . quotemeta($main::s_d_separator) . '\+-]+';
        my $regex2 = qr{ ^ $regex1 $ }x;

        #Pruefen ob nicht erlaubte Zeichen enthalten sind
        if( $s_tmp_data1 !~ $regex2 or $s_tmp_data2 !~ $regex2) {
          #wenigstens einer der beiden Werte ist nicht numerisch, sollte es aber sein

          $s_tmp_cmp_res = 1;
          $main::s_compare_message  = sprintf ("  Warning: column $s_tmp_field_no should be compared numerical, but the values aren't numerical!\n");
          $main::s_compare_message .= sprintf ("    It could help to provide a decimal separator.\n") if (!$main::s_d_separator);
          $main::s_compare_message .= sprintf ("    It could help to provide a 1000 separator.\n") if (!$main::s_t_separator);

          #Ungueltige(s) Zeichen ausgeben
          my ($s_not_valid_char1, $s_not_valid_char2, $s_not_valid_char);
          ($s_not_valid_char1 = $s_tmp_data1) =~ s/$regex1//xg;
          ($s_not_valid_char2 = $s_tmp_data2) =~ s/$regex1//xg;
          $s_not_valid_char = $s_not_valid_char1 . $s_not_valid_char2;
          my %seen;
          $s_not_valid_char =~ s/(.)/$seen{$1}++ ? "" : $1/xgse;
          $main::s_compare_message .= sprintf ("    Following character(s) occured: \"$s_not_valid_char\", valid ones are \"+-0123456789$main::s_t_separator$main::s_d_separator\"!\n");

        }

        else {
          #Tausendertrennzeichen "entsorgen" (temporaer)
          s/(\d)\Q$main::s_t_separator\E(?=\d{3})/$1/xg for ($s_tmp_data1, $s_tmp_data2);
          #Dezimaltrenner auf Perl internes Format umbiegen (temporaer)
          s/\Q$main::s_d_separator\E/\./x for ($s_tmp_data1, $s_tmp_data2);
          #Sollte das Vorzeichen hinter der Zahl stehen, wird es hier nach
          #vorne geholt (temporaer)
          s/(.*)([-\+])$/$2$1/x for ($s_tmp_data1, $s_tmp_data2);

          #beide Werte sind numerisch
          if(looks_like_number($s_tmp_data1) and looks_like_number($s_tmp_data2)) {
            #Vergleiche Spalte(n) numerisch
            if(($s_tmp_data1 - $s_tmp_data2) == 0) {
              #Werte sind numerisch identisch
              $s_tmp_cmp_res = 0;
            }
            else {
              #Werte sind numerisch nicht identisch
              $s_tmp_cmp_res = 1;
            }
          }
          #beide Felder sind leer
          elsif($s_tmp_data1 eq "" and $s_tmp_data2 eq "") {
            #Werte sind numerisch identisch
            $s_tmp_cmp_res = 0;
          }
          else {
            print STDERR "This should never occour, something ist wrong with > $s_tmp_data1 <\n";
            print STDERR "or > $s_tmp_data2 <! => exit 1\n";
            exit 1;
          }
        }
      }
      #die aktuelle Spalte soll nicht numerisch verglichen werden!
      else {
        ; # String Vergleich ist schon gemacht, siehe oben
      }
    }
    #es gibt keine Spalten die numerisch verglichen werden sollen!
    else {
      ; #der String Vergleich ist schon gemacht, siehe oben
    }

  }

  return $s_tmp_cmp_res;
}
# === Ende von Funktion cmp_data ==============================================

# === Funktion ================================================================
# Name:           diff_pos
# Beschreibung:   Position ermitteln, an der der erste Unterschied auftritt,
#                 diese Funktion ruft sich selber solange selber auf, bis der
#                 erste Unterschied gefunden ist.
# Beispiel:       $stelle = diff_pos($var1,$var2,$von,$bis)
# Parameter:      $var1 = erste Zeichenkette der "untersucht" werden soll
#                 $var2 = zweite Zeichenkette der "untersucht" werden soll
#                 $von  = ab dieser Position wird nach Unterschieden gesucht
#                 $bis  = bis zu dieser Position wird nach Unterschieden gesucht
# Rueckgabe Wert: Die Position an der der erste Unterschied aufgetreten ist,
#                 die Zaehlung beginnt auch hier bei 1!
# =============================================================================
#Ermitteln ab welcher Stelle sich zwei Strings unterscheiden.
#Das ganze funktioniert rekursiv, zuerst wird die "erste Haelfte" der Strings
#untersucht. Sind diese Teilstrings gleich, wird die zweite Haelfte miteinander
#verglichen, sind sie nicht gleich wird die "erste Haelfte" geteilt, und
#wiederum die erste haelfte hiervon untersucht, ...
#Da diese Funktion nur(!) aufgerufen wird, wenn zuvor festgestellt wurde, das
#sich die Zeichenketten voneinander unterscheiden, muss also die Position
#gefundenwerden.
sub diff_pos {
  my($ist, $soll, $von, $bis) = @_;

  if($main::s_ignore) {
    # Gross - Kleinschreibung ignorieren
    $ist  = lc($ist);
    $soll = lc($soll);
  }

  #zu untersuchenden Teil aus den Zeichenketten extrahieren
  my $anz_zeichen = $bis - $von + 1;
  my $ist_teil  = substr($ist ,$von,$anz_zeichen);
  my $soll_teil = substr($soll,$von,$anz_zeichen);

  my $neu_von = 0;
  my $neu_bis = 0;

  # Die Zeichenketten sind nicht gleich
  if ($ist_teil ne $soll_teil){
    #print "ne fuer Bereich";
    $neu_von = $von;
    $neu_bis = $von + int(($anz_zeichen)/2) - 1;
  }
  # Die Zeichenketten sind gleich
  else {
    #print "eq fuer Bereich ";
    $neu_von = $bis+1;
    $neu_bis = int($bis + 1 + ($bis/2));
    #aufpassen das die "bis" Position nicht hinter dem Ende der Zeichenkette liegt
    if ($neu_bis gt length($ist)) {
      $neu_bis = length($ist);
    }
    #was soll das ?????
    if ($von eq $bis) {
      $von++;
    }
  }

#  diff_pos($ist,$soll,$neu_von,$neu_bis) if(($bis - $von) != 0);
#  return ($von + 1);
  
  #print " von=$von bis=$bis  anz_zeichen=$anz_zeichen Bereich im soll = $soll_teil\n";

#  # rueckgabe der Stelle
#  if(($bis - $von) == 0) {
#    return ($von + 1);
#  }
#  # erneute Rekursion
#  else {
#    diff_pos($ist,$soll,$neu_von,$neu_bis);
#  }
  return ($von + 1) if(($bis - $von) == 0);

  # erneute Rekursion
  return diff_pos($ist,$soll,$neu_von,$neu_bis);

}
# === Ende von Funktion diff_pos ==============================================

# === Funktion ================================================================
# Name:           call_diff_pos
# Beschreibung:   Wraper Funktion fuer "diff_pos" um die Position des ersten
#                 Unterschieds zu ermitteln, und anzuzeigen
# Beispiel:       call_diff_pos($var1, $var2);
# Parameter:      $var1 = erste Zeichenkette die verglichen werden soll
#                 $var2 = zweite Zeichenkette die verglichen werden soll
# Rueckgabe Wert: -
# =============================================================================
sub call_diff_pos {
  my ($tmp_ist, $tmp_soll) = @_;

  my $ist_laenge  = length($tmp_ist);
  my $soll_laenge = length($tmp_soll);

  #die Position des ersten Unterschieds wird nur ermittelt wenn wenigstens eine
  #der beiden Zeichenketten mindestens $show_pos_length Zeichen lang ist, und
  #$show_pos_length selber groesser als 0 ist
  if (($ist_laenge >= $main::show_pos_length or $soll_laenge >= $main::show_pos_length) and $main::show_pos_length > 0) {
    my $bis = 0;
    my $von = 0;
    #ermitteln bis zu welchem Zeichen dier Haelfte der laengeren Zeichenkette
    #geht
    if ($ist_laenge gt $soll_laenge) {
      $bis = int($ist_laenge/2);
    }
    else {
      $bis = int($soll_laenge/2);
    }

    my $stelle = diff_pos($tmp_ist,$tmp_soll,$von,$bis); #Ermitteln an wo der erste Unterschied besteht
    my $first_difference = "  Position = ";
    for (my $ind = 1; $ind < $stelle; $ind++) {
      $first_difference = $first_difference . " ";
    }

    if($main::coloured_output){
      #$show_pos_length=1;  #da hier die ganze Zeile betrachtet wird, wird die
        #Position immer angezeigt
      $first_difference = $first_difference . " " x 7; # Da der Aufbau des
        # Outputs hier andes ist, muessen noch 7 Leerzeichen eingefuegt werden
    }

    $first_difference = $first_difference . "^";
    print "$first_difference\n";
  }
  #Ausgeben/Anzeigen an welcher Stelle der erste Unterschied ist -Ende-
  return 1;
}
# === Ende von Funktion call_diff_pos =========================================

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

csvdiff_function - collection of all subroutines defined for csvdiff.pl

=head1 SYNOPSIS

  use csvdiff_function;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for csvdiff_function, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Roland Schmitz, E<lt>roli@E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010 by Roland Schmitz

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.1 or,
at your option, any later version of Perl 5 you may have available.


=cut
